# -*- coding: utf-8 -*-
"""Utils for interfacing between pandas, pandera.SchemaModel and pyspark."""


from typing import Dict, List, Union

import pandas as pd
import pyspark
from pandera import DataFrameSchema
from pyspark.sql import SparkSession
from pyspark.sql.types import (
    BooleanType,
    DataType,
    DoubleType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

"""mapping from data type names to pyspark field types"""
TRANSFORM_DICT = {
    "int32": IntegerType(),
    "int64": LongType(),
    "Int64": LongType(),
    "float64": DoubleType(),
    "str": StringType(),
    "datetime64[ns]": TimestampType(),
    "bool": BooleanType(),
    "datetime64[ns, UTC]": TimestampType(),
    "datetime.date": DataType,
}


def create_spark_dataframe_using_pandera_model(
    cls: DataFrameSchema,
    spark: SparkSession,
    df: Union[pd.DataFrame, Dict, List[Dict]],
) -> pyspark.sql.DataFrame:
    """It converts the input into a spark dataframe following the class schema. Accept dictionaries, lists or dataframes as input.

    Args:
        spark (SparkSession): SparkSession
        df (Union[pd.DataFrame,Dict, List[Dict]]): Input data

    Returns:
        pyspark.sql.DataFrame: Spark dataframe with the class schema
    """
    # parse and validate the input
    if isinstance(df, dict) or isinstance(df, list):
        df = parse_dataframe_from_dict(cls, df)
    else:
        df = parse_dataframe_using_pandera_model(cls, df)
    return spark.createDataFrame(
        df, schema=pandera_model_to_spark_structype(cls)
    )


def parse_dataframe_from_dict(
    cls: DataFrameSchema, data: Union[List[Dict], Dict], **kwargs
) -> pd.DataFrame:
    """
    Initializes dataframe from dict. Wrapper for pd.DataFrame() that adds the
    dtype=object argument to avoid unasked type-castings.

    After creating the dataframe, it is parsed (with the parse method), according
    to the schema.

    Args:
        data (Union[List[Dict], Dict]): Dict, List of Dicts, or actually whatever
        input pd.DataFrame wouldn't sneeze at. In extra kwargs are required for
        compatibility, bring on the extra arguments, not gonna cry about that.

    Returns:
        pd.DataFrame: DataFrame post parsing.
    """

    df = pd.DataFrame(data=data, dtype=object, **kwargs)
    return parse_dataframe_using_pandera_model(cls, df)


def parse_dataframe_using_pandera_model(
    cls: DataFrameSchema, df: pd.DataFrame
) -> pd.DataFrame:
    """
    Parses DataFrame, adding required columns and doing valid type conversions
    according to schema typing. Includes a validation

    Args:
        df (pd.DataFrame): pandas DataFrame

    Returns:
        pd.DataFrame: pandas DataFrame with correct types according to schema.
    """
    # Add missing columns to dataframe as empty columns
    df_cols = df.columns
    for column in cls.columns.values():
        if column.name not in df_cols and not column.regex:
            df[column.name] = pd.Series(dtype=str(column.dtype))

    # Validate types and convert
    df = cls.validate(df)
    return df


def pandera_model_to_spark_structype(
    cls: DataFrameSchema, df_cols: List[str] = None
) -> StructType:
    """
    Returns spark StructType with schema or corresponding Pandera schema.

    Args:
        df_cols (Optional[List[str]], optional): List of DataFrame col names of corresponding schema,
        useful for returning StructType with a particular column order.
        Defaults to None. If None, uses default schema column ordering.

    Returns:
        StructType: schema for pyspark DataFrame
    """

    schema_cols = cls.columns

    if df_cols is None:
        df_cols = [col for col in schema_cols.keys()]

    struct = StructType()
    for col in df_cols:
        struct.add(
            StructField(
                col,
                TRANSFORM_DICT[str(schema_cols.get(col).dtype)],
                schema_cols.get(col).nullable,
            )
        )
    return struct



# -*- coding: utf-8 -*-
"""Spark applyInPandas Pipeline class"""

from typing import List, Optional, Union

from pandas import DataFrame as PandasDataFrame
from pandera import DataFrameSchema
from pyspark.sql import DataFrame as SparkDataFrame
from typeguard import typechecked

from pysparkpipe.exc import (
    InputLayerMissingGroupingColsError,
    LayersConsistencyError,
    PipelineCompileError,
)
from pysparkpipe.pipeline.layers import Layer
from pysparkpipe.pipeline.utils import pandera_model_to_spark_structype


class Pipeline:
    """Pipeline class. A pipeline is a sequence of layers that are applied to the data grouped by the grouping columns."""

    @typechecked
    def __init__(
        self,
        grouping_cols: List[str],
        validate_inputs: bool = True,
        validate_outputs: bool = True,
    ) -> None:
        """Pipeline constructor.

        Args:
            grouping_cols (List[str]): List of grouping columns.
            validate_inputs (bool, optional): If True, by default the inner Layers will validate the input dataframes. Defaults to True.
            validate_outputs (bool, optional): If True, by defualt the inner Layers will validate the output dataframes. Defaults to True.

        Raises:
            ValueError: If grouping_cols is not a list of strings.
            ValueError: If validate_inputs is not a boolean.
            ValueError: If validate_outputs is not a boolean.
        """

        self.grouping_cols = grouping_cols  # type: List[str]
        self.layers = []  # type: List[Layer]
        self.validate_inputs = validate_inputs  # type: bool
        self.validate_outputs = validate_outputs  # type: bool

    def add_layer(
        self,
        new_layer: Layer,
    ):
        # assert new_layer is consistent with previous layers
        if len(self.layers) > 0:
            if not new_layer.is_consistent_with(self.layers[-1]):
                raise LayersConsistencyError(
                    new_layer_name=new_layer.name,
                    prev_layer_name=self.layers[-1].name,
                )
        else:
            if not isinstance(new_layer.input_schema, list):
                sch_lst = [new_layer.input_schema]
            else:
                sch_lst = new_layer.input_schema
            for s in sch_lst:
                if not set(self.grouping_cols).issubset(set(s.columns.keys())):
                    raise InputLayerMissingGroupingColsError(
                        grouping_cols=self.grouping_cols, schema=s
                    )
        # if consistent, add the new layer to pipeline
        self.layers.append(new_layer)

    def add(
        self,
        transform: callable,
        input_schema: DataFrameSchema,
        output_schema: DataFrameSchema,
        validate_input: Optional[bool] = None,
        validate_output: Optional[bool] = None,
    ) -> None:
        # if validate_input is None, use the pipeline's validate_inputs
        if validate_input is None:
            validate_input = self.validate_inputs
        # if validate_output is None, use the pipeline's validate_outputs
        if validate_output is None:
            validate_output = self.validate_outputs

        # create a new layer
        new_layer = Layer(
            transform=transform,
            input_schema=input_schema,
            output_schema=output_schema,
            validate_input=validate_input,
            validate_output=validate_output,
        )
        # add the new layer to pipeline
        self.add_layer(new_layer)

    def compile(
        self,
        df: Optional[Union[SparkDataFrame, PandasDataFrame]] = None,
        df2: Optional[SparkDataFrame] = None,
    ) -> None:
        """Compile the pipeline."""

        if df2 is not None and len(self.input_schema) != 2:
            raise PipelineCompileError(
                msg="Pipeline must have exactly two input schemas to be compiled with two input dataframes."
            )
        elif df2 is None and isinstance(self.input_schema, list):
            raise PipelineCompileError(
                msg="Pipeline must have exactly one input schema to be compiled with one input dataframe."
            )

        for d in [df, df2]:
            if d is not None:
                # assert that df has grouping_cols
                if not all(col in df.columns for col in self.grouping_cols):
                    PipelineCompileError(
                        f"Input dataframe must have all the grouping columns. Got : {df.columns}"
                    )
        # assert that self.layers is not empty
        if not self.layers:
            raise PipelineCompileError(
                msg="Pipeline must have at least one layer."
            )
        if not isinstance(self.output_schema, DataFrameSchema):
            raise PipelineCompileError(
                msg="Output schema cannot be mutiple. pyspark's applyInPandas only supports single output."
            )

    @property
    def input_schema(self) -> DataFrameSchema:
        """Input schema of the pipeline."""
        if len(self.layers) == 0:
            return None
        return self.layers[0].input_schema

    @property
    def output_schema(self) -> DataFrameSchema:
        """Output schema of the pipeline."""
        if len(self.layers) == 0:
            return None
        return self.layers[-1].output_schema

    def fit(
        self, df: Union[PandasDataFrame, List[PandasDataFrame]]
    ) -> PandasDataFrame:
        """Fit the pipeline to the data. Apply all layers to the data to a single group."""
        for layer in self.layers:
            df = layer.transform(df)
        return df

    @typechecked
    def apply_in_pandas(self, df: PandasDataFrame) -> PandasDataFrame:
        """Apply the pipeline to the data using pandas 'groupby'. Apply all layers to the data grouped by the grouping columns.

        Args:
            df (DataFrame): DataFrame to apply the pipeline to.

        Returns:
            DataFrame: DataFrame with the pipeline applied.

        Raises:
            TypeCheckError: If df is not a DataFrame.
            PipelineCompileError: If the pipeline could not be compiled.
        """
        self.compile(df)
        grouped_df = df.groupby(self.grouping_cols)
        return grouped_df.apply(self.fit)

    @typechecked
    def apply_in_spark(
        self, df: SparkDataFrame, df2: Optional[SparkDataFrame] = None
    ) -> SparkDataFrame:
        """Apply the pipeline to the data using Spark 'groupBy'. Apply all layers to the data grouped by the grouping columns.
        In the case of multiple Spark DataFrames, the pipeline is applied in a cogroup fashion.

        Args:
            df (SparkDataFrame): Spark DataFrame to apply the pipeline to, or a list of Spark DataFrames.

        Returns:
            SparkDataFrame: Spark DataFrame with the pipeline applied.

        Raises:
            TypeCheckError: If df is not a SparkDataFrame.
            PipelineCompileError: If the pipeline could not be compiled.
        """

        self.compile(df, df2)

        grouped_df = df.groupby(self.grouping_cols)
        if df2 is not None:
            grouped_df = grouped_df.cogroup(df2.groupby(self.grouping_cols))

            def to_apply(key, left, right):
                return self.fit([left, right])

        else:

            def to_apply(key, df):
                return self.fit(df)

        return grouped_df.applyInPandas(
            to_apply,
            schema=pandera_model_to_spark_structype(self.output_schema),
        )



# -*- coding: utf-8 -*-
"""Layer class for the pipeline."""

from typing import List, Union

from pandas import DataFrame as PandasDataFrame
from pandera import DataFrameSchema
from typeguard import TypeCheckError, typechecked

from .utils import parse_dataframe_using_pandera_model


class Layer:
    """Layer class for the pipeline.

    Args:
        transform (callable): Transform function.
        grouping_cols (List[str]): List of columns to group by.
        input_schema (Union[DataFrameSchema, List[DataFrameSchema]]): Input schema(s), must contain grouping_cols.
        output_schema (Union[DataFrameSchema, List[DataFrameSchema]]): Output schema(s), must contain grouping_cols.
        validate_input (bool, optional): Validate input schema. Defaults to True.
        validate_output (bool, optional): Validate output schema. Defaults to True.

    Returns:
        None

    Raises:
        TypeError: If transform is not callable.
        TypeError: If input_schema is not instance of DataFrameSchema.
        TypeError: If output_schema is not instance of DataFrameSchema.
        TypeError: If grouping_cols is not subset of input_schema columns.
        TypeError: If grouping_cols is not subset of output_schema columns.
        TypeError: If validate_input is not bool.
        TypeError: If validate_output is not bool.

    """

    @typechecked
    def __init__(
        self,
        transform: callable,
        input_schema: Union[DataFrameSchema, List[DataFrameSchema]],
        output_schema: Union[DataFrameSchema, List[DataFrameSchema]],
        validate_input: bool = True,
        validate_output: bool = True,
    ) -> None:
        """Initialize the layer."""
        self._transform = transform
        self.input_schema = input_schema
        self.output_schema = output_schema
        self.validate_input = validate_input
        self.validate_output = validate_output

        # assert that transform is callable
        if not callable(self._transform):
            raise TypeCheckError("Transform function must be callable.")

        self.name = self._transform.__name__  # type: str

    def validate(
        self,
        df: Union[PandasDataFrame, List[PandasDataFrame]],
        schema: Union[DataFrameSchema, List[DataFrameSchema]],
    ) -> Union[PandasDataFrame, List[PandasDataFrame]]:
        """Validate a pandas DataFrame (or a list of pandas DataFrames) against schema (or list of schemas.)

        Args:
            df (Union[PandasDataFrame, List[PandasDataFrame]]): DataFrame to validate.
            schema (Union[DataFrameSchema, List[DataFrameSchema]]): Schema to validate against.

        Returns:
            Union[PandasDataFrame, List[PandasDataFrame]]: Validated DataFrame (or list of validated DataFrames).
        """
        if isinstance(df, list):
            # if df is a list of dataframes, validate each dataframe and return list of validated dataframes
            return [self.validate(d, s) for d, s in zip(df, schema)]
        return parse_dataframe_using_pandera_model(schema, df)

    def transform(
        self, df: Union[PandasDataFrame, List[PandasDataFrame]]
    ) -> Union[PandasDataFrame, List[PandasDataFrame]]:
        """Transform a pandas DataFrame.

        Args:
            df (DataFrame): DataFrame to transform.

        Returns:
            DataFrame: Transformed DataFrame.
        """
        if self.validate_input:
            # validate input dataframe against input schema
            df = self.validate(df=df, schema=self.input_schema)

        if isinstance(df, list):
            # if df is a list of dataframes, unpack list
            df = self._transform(*df)
        else:
            df = self._transform(df)

        if self.validate_output:
            # validate output dataframe against output schema
            df = self.validate(df=df, schema=self.output_schema)
        return df

    def is_consistent_with(self, other: "Layer") -> bool:
        """Check if layer is consistent with other layer, i.e. if they can be chained in a pipeline (the new layer is applied after the other layer).

        Args:
            other (Layer): Other (previous) layer to check consistency with.

        Returns:
            bool: True if consistent, False otherwise.
        """
        if self.input_schema != other.output_schema:
            return False
        return True



# -*- coding: utf-8 -*-
"""Exceptions handling."""

import abc
from contextlib import contextmanager
from typing import Any, Generator, Type, Union


class PySparkPipeBaseException(abc.ABC, BaseException):
    """Base class for custom errors and exceptions.
    Example:
        >>> class MyError(PySparkPipeBaseException):
                msg_template = "Value ``{value}`` could not be found"
        >>> raise MyError(value="can't touch this")
        (...)
        MyError: Value `can't touch this` could not be found

    """

    @property
    @abc.abstractmethod
    def msg_template(self) -> str:
        """Un template para imprimir cuando una excepción es levantada.

        Example:
            "Value ``{value}`` is not found "

        """

    def __init__(self, **ctx: Any) -> None:
        self.ctx = ctx
        super().__init__()

    def __str__(self) -> str:
        txt = self.msg_template
        for name, value in self.ctx.items():
            txt = txt.replace("{" + name + "}", str(value))
        txt = txt.replace("`{", "").replace("}`", "")
        return txt


@contextmanager
def change_exception(
    raise_exc: Union[PySparkPipeBaseException, Type[PySparkPipeBaseException]],
    *except_types: Type[BaseException],
) -> Generator[None, None, None]:
    """Context Manager to replace exceptions with custom ones.

    See also:
        :func:`pydantic.utils.change_exception`

    """
    try:
        yield
    except except_types as exception:
        raise raise_exc from exception  # type: ignore


class EnvVarNotFound(PySparkPipeBaseException, NameError):
    """Raise this when a table name has not been found."""

    msg_template = "Enviroment variable `{env_var}` can't be found"


class LayersConsistencyError(PySparkPipeBaseException, NameError):
    msg_template = "The new layer: `{new_layer_name}` is not consistent with the previous layer `{prev_layer_name}`"


class InputLayerMissingGroupingColsError(PySparkPipeBaseException, NameError):
    msg_template = "The new layer: `{layer_name}` is not consistent: grouping columns: `{grouping_cols}` must be subset of the schema columns: `{schema}`"


class PipelineCompileError(PySparkPipeBaseException, NameError):
    msg_template = "The pipeline can't be compiled: `{msg}`"
